
/*    in here is anything related to animating the anymails
 *
 *    every anymail has one controller which has three
 *    animators, one for the sprites and two for inner
 *    movement (offset to sceenposition) and inner rotation.
 *    once the controller is ticked, it passes that tick to
 *    each of it's animators. this syncronises the sprites
 *    with the inner movement.
 *
 *    each animator holds one pointer to a set of values or
 *    sprites. this is to lower memory usage.
 */
 
 
/*    interface AnimationStepper
 *
 *    implemented by SpriteAnimatior and FloatAnimator,
 *    a wrapper to use different types of animators in
 *    one controller.
 */
 
interface AnimationStepper
{
    void tick();
    void tick( int _t );
}


/*    class AnimationController
 *
 *    controlls multiple animators, 
 *    passes a tick to each of them for each step.
 */
 
class AnimationController
{
    AnimationStepper[] animations;
    int[] indices;
    
    AnimationController ( AnimationStepper[] _a )
    {
        animations = _a;
    }
    
    void tick () { tick( 1 ); }
    
    void tick ( int _a )
    {
        int i=animations.length-1; try { do {
            
            animations[i].tick(_a);
            
        i--; } while(true); } catch (ArrayIndexOutOfBoundsException aioobe) {;}
    }
}


/*    class FloatAnimator
 *
 *    steps thru a bunch of float values,
 *    used for internal rotations and movements
 */

class FloatAnimator
implements AnimationStepper
{
    FloatContainer container;
    int index;
    float min_value =  1000;
    float max_value = -1000;
    
    FloatAnimator ( ) { }
    
    FloatAnimator ( FloatContainer _cont )
    {
        container = _cont;
        index = (int)(random(container.values.length-1));
        reset();
    }
    
    void reset ()
    {
        for ( int i=0; i < container.values.length; i++ )
        {
            min_value = min( min_value, container.values[i] );
            max_value = max( max_value, container.values[i] );
        }
    }
    
    void tick ( ) { tick(1); }
    
    void tick ( int _t )
    {
        index+=_t; index %= container.values.length;
    }
    
    float value ()
    {
        return container.values[index];   
    }
}


/*    class SpriteAnimator
 *
 *    steps thru a bunch of PImages,
 *    used for sprite animations
 */

class SpriteAnimator
implements AnimationStepper
{
    SpriteContainer container;
    int index;
    
    SpriteAnimator ( SpriteContainer _cont )
    {
        container = _cont;
        index = (int)(random(container.images.length-1));
    }
    
    void tick ( ) { tick(1); }
    
    void tick ( int _t )
    {
        index+=_t; index %= container.images.length;   
    }
    
    PImage image ()
    {
        return container.images[index];
    }
    
    PImage next ()
    {
        tick();
        return container.images[index];
    }
}


/*    class FloatContainer
 *
 *    houses a couple of floats,
 *    used to ease memory usage
 *    ( one species shares one FloatContainer )
 */

class FloatContainer
{
    float[] values;
    
    FloatContainer ( float[] _vals )
    {
        values = _vals;   
    }
}


/*    class SpriteContainer
 *
 *    houses a couple of PImages,
 *    used to ease memory usage
 *    ( one species shares one SpriteContainer )
 */

class SpriteContainer
{
    PImage[] images;
 
    SpriteContainer ( PImage[] _imgs )
    {
        images = _imgs;
    }
}


 
/*
 *    This work is licensed under a 
 *    Creative Commons Attribution-Noncommercial-Share Alike 3.0 License.
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/
 */
