
/*    the anymails
 *
 *    an anymail is a bundle that links an Emlx (Email)
 *    on the data-side to a screen representation on the
 *    user-side.
 */
 
boolean showInfo = false;

class Anymail
implements MausDing, Comparable
{
    Emlx email;
    Species species;
    int groupID;
    
    String name; // override the name in Emlx, is loaded from .addresses
    
    final static int STATE_UNREAD   = 0;
    final static int STATE_READ     = 1;
    final static int STATE_ANSWERED = 2;
    
    final static int AGE_1 = 4;
    final static int AGE_2 = 3;
    final static int AGE_3 = 2;
    final static int AGE_4 = 1;
    final static int AGE_5 = 0;
    
    final static float DAMPING = 2.0f;
    
    // propagating calls in the chain
    final static int THIS_ONLY = -2;
    final static int CHAIN     =  0;
    final static int UPWARDS   =  1;
    final static int DOWNWARDS = -1;
    
    final static boolean CLEAR = true;
    final static boolean DRAW = false;
    
    int state = -1;
    int age = -1;
    int type = -1;
    
    boolean dirty = true;
    
    Vector3D loc;
    Vector3D vel;
    Vector3D acc;
    
    float wandertheta;
    float maxforce, maxforceDefault;    // Maximum steering force
    float maxspeed, maxspeedDefault;    // Maximum speed
  
    float wanderR = 5.0f;  // steering radius of circle to find new direction on
    float wanderD = 70.0f; // circle distance to loc.x, loc.y
    float change  = 0.30f; // max radians to change direction per cycle
    
    Anymail target;   // daisy-chaining, ahead
    Anymail follower; // daisy-chaining, behind
    
    int xHashKey;    // these are set in GridLookUp.add() and 
    int yHashKey;    // contain the grid-indices this Anymail is in

    float xo, yo;   // previous coords
    float x, y;     // screen coords
    int iw,  ih;    // width, height of current sprite
    int iw2, ih2;   // iw, ih times two
    int imax;       // max(iw,ih)
    int imax2;      // imax / 2
    
    Area area;
    
    //boolean showInfo        = false;    // info-box
    boolean highLight       = false;    // beige circle
    boolean isSelected      = false;    // mouse-over, white circle
    
    Anymail ( Species _species, float _x, float _y )
    {        
        species = _species;

        iw = species.sprite().width;
        iw2 = iw/2;
        ih = species.sprite().height;
        ih2 = ih/2;
        
        acc = new Vector3D(0,0,0);
        vel = new Vector3D(0,0,0);
        loc = new Vector3D(_x,_y,0);
        
        neighbor = new Vector3D(0,0,0);
        
        wandertheta = radians(random(360));

        maxspeed = 1.0;
        maxforceDefault = 0.8;
        maxforce = maxforceDefault;
    }
    
    void setArea ( Area _area )
    {
        area = _area;
        area.registerMausDing( this );
    }
    
    void setState ( int _state )
    {
        species.setState( _state );
        state = species.state;
        reset();
    }
    
    void setAge ( long start, long end )
    {
        long emaildate = email.date_received - (start/(long)1000);
        long datedist = (end - start)/(long)1000;
        float val = ( 1.0 / (float)datedist ) * (float)emaildate;
        setAge( val );
    }
    
    void setAge ( float _val )
    {
        species.setAge( _val );
        age = species.age;
        reset();
    }
    
    void setHighlight ( boolean on, int deep )
    {
        if ( deep != THIS_ONLY )
        {
            if ( deep >= CHAIN && target != null )
                target.setHighlight ( on, UPWARDS );
            if ( deep <= CHAIN && follower != null )
                follower.setHighlight ( on, DOWNWARDS );
        }
        
        highLight = on;
        
        if ( !highLight )
            reset();
        else if ( target != null )
        {
            maxspeed = target.maxspeed;
            maxforce = target.maxforce;   
        }
    }
    
    void chain ( Anymail _target )
    {
        target = _target;
        target.follower = this;
    }
    
    void unChain ( )
    {
        if ( target != null
             && target.follower != null
             && target.follower.equals(this)
           )
           target.follower = null;
        target = null;
        follower = null;
        
        reset();
    }
    
    boolean flipped = false;
    
    void flipChain ( int deep )
    {
        if ( deep != THIS_ONLY )
        {
            if ( deep >= CHAIN && target != null )
                target.flipChain ( UPWARDS );
            if ( deep <= CHAIN && follower != null )
                follower.flipChain ( DOWNWARDS );
        }
        
        flipped = !flipped;
    }
    
    void reset ()
    {
        iw = species.sprite().width;
        iw2 = iw/2;
        ih = species.sprite().height;
        ih2 = ih/2;
        
        imax = max(iw,ih)+5;
        imax2 = imax / 2;
        
        if ( neighbor == null ) neighbor = new Vector3D(0,0,0);
        
        maxspeedDefault =
            state == Anymail.STATE_ANSWERED
                     ? Speeds.SLOW
                     : (state == Anymail.STATE_READ
                        ? Speeds.MEDIUM
                        : Speeds.FAST);
        maxspeed = maxspeedDefault;
        wanderD = maxspeed * 80.0;
        wanderR = maxspeed * 7.0;
        
        targetDist = 0.0;
        
        dirty = true;
    }
    
    // ANYMAIL // D R A W I N G
    
    void clear ()
    {
        if ( dirty ) return;
        draw( CLEAR );
        dirty = true;
    }

    void draw ()
    {
        draw( DRAW );
        dirty = false;
    }
    
    void draw ( boolean clear )
    {
        
        if ( DEBUG ) // show forces, target, ...
        {
            stroke(Colors.BLACK); strokeWeight( 1 ); noFill();
            line( loc.x, loc.y, circleloc.x, circleloc.y );
            ellipse( circleloc.x, circleloc.y, wanderR*2, wanderR*2 );
            line( circleloc.x, circleloc.y, _target.x, _target.y );
            
            stroke(0xFF00FF00);
            rect(loc.x+x-imax2, loc.y+y-imax2, imax, imax);
            /*
            Anymail[] neighbors = gridLookUp.get( this );
            if ( neighbors != null && neighbors.length != 0 )
            {
                for ( int i=0; i< neighbors.length; i++ )
                {
                    if (    neighbors[i] == this
                         //|| neighbors[i].groupID == groupID
                         || (target != null && neighbors[i] == target)
                         || (follower != null && neighbors[i] == follower)
                    ) continue;
                    beginShape(LINES);
                        stroke(0xFFFF00FF);
                        vertex( loc.x+3, loc.y+3);
                        stroke(0xFFFFFF00);
                        vertex(neighbors[i].loc.x-3, neighbors[i].loc.y-3 );
                    endShape();
                }
            }
            */
            noStroke();
        }
            
        noStroke();
        pushMatrix();
            translate( loc.x, loc.y );
            rotate( vel.heading2D() + species.rotation() );
            translate( species.movement(), 0 );
        
            fill( clear ? Colors.BACKGROUND : 0xFFFFFFFF );
            beginShape();
                if ( !clear ) {
                    texture(species.sprite());
                    vertex( -iw2, -ih2,  0,  0);
                    vertex(  iw2, -ih2, iw,  0);
                    vertex(  iw2,  ih2, iw, ih);
                    vertex( -iw2,  ih2,  0, ih);
                } else {
                    vertex( -iw2, -ih2 );
                    vertex(  iw2, -ih2 );
                    vertex(  iw2,  ih2 );
                    vertex( -iw2,  ih2 );
                }
            endShape();
        popMatrix();
        
        if ( !clear ) dirty = false;
    }
    
    void drawHighlight ( )
    {
        drawHighlight ( CHAIN, false );
    }
    
    void drawHighlight ( boolean clear )
    {
        drawHighlight( CHAIN, clear );
    }
    
    float targetDist = 0.0;
    
    void drawHighlight ( int deep, boolean clear )
    {
        //if ( true ) return;
        
        if ( deep != THIS_ONLY )
        {
            // pass on up and down the chain
            if ( deep >= CHAIN && target != null)
                target.drawHighlight(UPWARDS, clear);
            if ( deep <= CHAIN && follower != null)
                follower.drawHighlight(DOWNWARDS, clear);
        }
        
        // draw conection to target
        
        if ( (!flipped && target != null) || (flipped && follower != null) )
        {
            if ( targetDist < 300.0 )
            {
                stroke( clear ? Colors.BACKGROUND : Colors.BEIGE_DARK );
                strokeWeight( targetDist > 150 ? min((int)((300.0)/(targetDist-150)),STROKE_WEIGHT) : STROKE_WEIGHT );
                //line( loc.x,loc.y,target.loc.x,target.loc.y );
                if ( !flipped )
                    thickLine( loc.x, loc.y, target.loc.x, target.loc.y );
                else
                    thickLine( loc.x, loc.y, follower.loc.x, follower.loc.y );
                strokeWeight( 1 );
            }
        }
        
        // draw ellipse
        
        noStroke();
        pushMatrix();
            translate( loc.x, loc.y );
            pushMatrix();
                rotate( vel.heading2D() + species.rotation() );
                translate( species.movement(), 0 );
        
                fill( Colors.BEIGE_DARK );
                ellipse(0,0,imax,imax);
                if ( isSelected )
                {
                    fill( Colors.WHITE );
                    ellipse(0,0,imax-6,imax-6);
                }
            popMatrix();
        popMatrix();
    }
    
    void drawInfo ()
    {
        if ( !showInfo || !isSelected ) return;
                
        // this could be less jumpy ...
        // currently we only have 4 possible positions,
        // could be always positioned towards center and 
        // follow the anymail with a slight dumping ...
        
        int swidth = 213;
        int sx = loc.x < area.xwidthHalf ? iw+2 : -iw-2- swidth ; sx += (int)loc.x;
        
        int sheight = 84;
        int sy = loc.y < area.yheightHalf ? -10 : -sheight+10 ; sy += (int)loc.y;
        
        shadowedBox.draw((int)sx,(int)sy,(int)swidth,(int)sheight);
    }
  
        // ANYMAIL //  M O V E M E N T
/*
 *    movement code originates from daniel shiffmans Vector3D examples here:
 *    http://www.shiffman.net/teaching/nature/steering/
 */
 
    void move ()
    {
        species.tick();            
        wander();
        dirty = true;
    }
    
    Vector3D _target;
    Vector3D circleloc;
    Vector3D neighbor;
    
    void wander ()
    {
        boolean slowdown = false;
        Anymail tTarget, tFollower;
        
        if ( !flipped )
        {
            tTarget   = target;
            tFollower = follower;
        }
        else
        {
            tTarget   = follower;
            tFollower = target;
        }
        
        if ( tTarget != null )
            targetDist = fastDist( loc.x, loc.y, tTarget.loc.x, tTarget.loc.y );
        
        //if ( highLight ) {
            
            Anymail[] neighbors = gridLookUp.get( this );
            
            if ( neighbors != null && neighbors.length != 0 )
            {
                for ( int i=0; i<neighbors.length; i++ )
                {
                    if (    neighbors[i] == this
                         //|| neighbors[i].groupID == groupID
                         || (tTarget   != null && neighbors[i] == tTarget)
                         || (tFollower != null && neighbors[i] == tFollower)
                         || (!highLight && !neighbors[i].highLight )
                    ) continue;
                    
                    if ( DEBUG ) {
                        beginShape(LINES);
                            stroke(0xFFFF00FF);
                            vertex( loc.x+3, loc.y+3);
                            stroke(0xFFFFFF00);
                            vertex(neighbors[i].loc.x-3, neighbors[i].loc.y-3 );
                        endShape();
                    }
                    
                    Vector3D offset = Vector3D.subtract(loc, neighbors[i].loc);
                    offset.multiplyBy( Forces.ANTIGRAVITY/offset.magnitude() );
                    
                    neighbor.add( offset );
                    
                    offset.multiplyBy( -2.0 );
                    
                    neighbors[i].neighbor.add( offset );
                }
            }
        //}
        
        if ( !isSelected )
        {
            Vector3D nLoc = loc.copy();
            nLoc.add(neighbor);
            if ( area.inside(nLoc.x-imax2,nLoc.y-imax2,imax,imax) )
            {
                loc = nLoc;
                neighbor.multiplyBy( Forces.ANITFRICTION );
            }
            else
                neighbor.setXYZ(0,0,0);
        }
        else // selected
        {
            neighbor.setXYZ(0,0,0);
            return;
        }
        
        if (  tTarget != null && highLight ) // daisy-chained
        {
            Vector3D targetB = Vector3D.subtract(loc,tTarget.loc);
                
            float tMag = targetB.magnitude();
            targetB.normalize();
                
            if ( tMag <= tTarget.imax )
            {
                float angle = targetB.heading2D()+PI;
                float magn = vel.magnitude();
                vel = new Vector3D(magn*cos(angle),magn*sin(angle),0);
                return;
            }
            else
            {
                targetB.multiplyBy(tTarget.imax);
                targetB.add(tTarget.loc);
            }
                
                _target = targetB;
            
                circleloc = _target;
                wandertheta = targetB.heading2D();
        }
        else  // wandering along oh so free
        {            
            wandertheta += random(-change,change);
            
            circleloc = vel.copy();
            circleloc.normalize();
            circleloc.multiplyBy(wanderD);
            circleloc.add(loc);
    
            Vector3D circleOffSet = new Vector3D(wanderR*cos(wandertheta),wanderR*sin(wandertheta),0);
            _target = Vector3D.add(circleloc,circleOffSet);
            
            int wanderR2 = (int)wanderR<<1;
            if ( !area51.inside(circleloc.x-wanderR,circleloc.y-wanderR,wanderR2,wanderR2) )
            {
                Vector3D center = new Vector3D(area51.x+area51.width/2.0, area51.y+area51.height/2.0, 0);
                Vector3D newLoc = Vector3D.subtract(center,_target.copy()); // a Vector3D from the center to _target
                wandertheta = newLoc.heading2D();
                wandertheta += random(-change,change);
                
                circleOffSet = new Vector3D(wanderR*cos(wandertheta),wanderR*sin(wandertheta), 0);
                _target = Vector3D.add(circleloc,circleOffSet);
            }
        }
        
        // steer -----
        
        Vector3D steer;
        Vector3D desired = Vector3D.subtract(_target,loc);
        float d = desired.magnitude();
        
        float speed   = highLight && targetDist >= (imax<<1) ? Speeds.PANIC : maxspeed;
        float force   = maxforce;
        
        if (d > 0) {
            
            desired.normalize();
            if ((slowdown) && (d < DAMPING))
                desired.multiplyBy(speed*(d / DAMPING));
            else
                desired.multiplyBy(speed);
            
            steer = Vector3D.subtract(desired,vel);
            steer.limit(force);
        
       } else
            steer = new Vector3D(0,0,0);
        
        acc.add( steer );
        
        // update ------
        
        vel.add(acc);
        vel.limit(highLight ? Speeds.PANIC : maxspeed);
        Vector3D newLoc = loc.copy();
        newLoc.add(vel);
        
        if ( area.inside(newLoc.x-imax2, newLoc.y-imax2, imax, imax) )
            loc = newLoc;
        else
        {
            if ( newLoc.x-imax2-2 > area.x && newLoc.x+imax+4 < area.xwidth )
                loc.x = newLoc.x;
            if ( newLoc.y-imax2-2 > area.y && newLoc.y+imax+4 < area.yheight )
                loc.y = newLoc.y;
        }
        
        acc.setXYZ(0,0,0);
    }
  
  // INTERFACE MausDing

    boolean check (int _mx, int _my) {
        if ( _mx < loc.x-iw2 || _mx > loc.x-iw2+iw || _my < loc.y-ih2 || _my > loc.y-ih2+ih ) return false;
        return true;
    }
	
    public void enter () {
        isSelected = true;
        selectedAnymail = this;
        
        setHighlight( true, CHAIN );
        flipChain( UPWARDS );
        
        if ( showInfo && isSelected ) shadowedBox.applyAnymail( this );
    }
    
    public void over () {
        isSelected = true;
        setHighlight( true, CHAIN );
    }
	
    public void exit () {
        isSelected = false;
        
        setHighlight( false, CHAIN );
        flipChain( UPWARDS );
        
        if ( selectedAnymail != null && selectedAnymail.equals(this) )
            selectedAnymail = null;
            
        clickcnt = 0; clickts = 0;
    }
    
    public void pressed ()
    {        
        xo = loc.x;
        yo = loc.y;
        
        long ts = System.currentTimeMillis() - clickts;
        if ( ts > 400 )
            clickcnt = 0;
    }

    long clickts = 0;
    int clickcnt = 0;
    public void released ()
    {
        long ts = System.currentTimeMillis() - clickts;
        if ( clickts > 0 && ts <= 400 && clickcnt == 1 ) doubleClicked();
        
        if ( clickcnt == 0 )
            clickts = System.currentTimeMillis();
            
        clickcnt++; clickcnt%=2;
    }
    
    public void doubleClicked ()
    {
        if ( true ) return; // just a test, would open the mail in Mail.app
        
        clickcnt = 0; clickts = 0;
        String folder = System.getProperty("user.home") + "/Library/Mail/";/*+
                                  applemailreader.mailboxes[
                                        applemailreader.mailboxMap.indexForValue( email.mailbox_id )
                                                        ].folderpath;*/
        String file = email.id + ".emlx";
        
        String filepath = AppleMailToolkit.shellExec( "find \""+folder+"\" -iname \""+file+"\" -print" ).get(0).toString();
        AppleMailToolkit.shellExecSilent( "open \""+filepath+"\"" );
        println( "Trying to open "+file+" in Mail." );
    }
    
    public void releasedOutside () {
    }
	
    public void dragged (int _dx, int _dy)
    {
        if ( area.inside(xo+_dx-imax2, yo+_dy-imax2, imax, imax) ) // can't be dragged outside
        {
            loc.x = xo + _dx;
            loc.y = yo + _dy;
        }
    }
	
    public void pressedRight () {
        showInfo = !showInfo;
        if ( showInfo ) shadowedBox.applyAnymail( this );
    }
	
    public void draggedRight (int _dx, int _dy) { }
	
    public void releasedRight () { }
	
    public void releasedOutsideRight () { }
    

    // ----- INTERFACE:  C O M P A R A B L E  ---------------
    
    public int compareTo ( Object _o )
    {
        /* From the JavaDocs:
         * Compares this object with the specified object for order.
         * Returns a negative integer, zero, or a positive integer as 
         * this object is less than, equal to, or greater than the 
         * specified object.
         *
         * But we don't just compare here, we implement a total ordering
         * dependent on the Grouping.XXX states .. juchuuhhh!
         */
        //if ( Grouping.BY_STATE == -1 || Grouping.ORDER_STATE == -1 ) return 0;
        
        Anymail other = (Anymail)_o;
        
        int retVal = compareGrouping( other );
        
        // same group, then apply ordering (arrangement)
        if ( retVal == 0 )
            retVal = compareOder( other );
        
        return retVal;
    }
    
    int compareGrouping( Anymail _a )
    {
        switch ( Grouping.BY_STATE )
        {
            case Grouping.BY_STATUS:
                return compareStatus( _a );
            case Grouping.BY_TIME:
                return compareAge( _a );
            case Grouping.BY_SPECIES:
                return compareSpecies( _a );
            case Grouping.BY_PERSON:
                return compareAddress( _a );
        }
        return 0;
    }
    
    int compareOder ( Anymail _a )
    {
        int retVal = 0;
        
        switch ( Grouping.ORDER_STATE )
            {
                case Grouping.ORDER_SIZE:
                    retVal = email.bytes > _a.email.bytes ? 1 : ( email.bytes == _a.email.bytes ? 0 : -1 );
                    break;
                case Grouping.ORDER_STATUS:
                    retVal = compareStatus( _a );
                    break;
                case Grouping.ORDER_TIME:
                    retVal = compareTime( _a );
                    break;
                case Grouping.ORDER_SPECIES:
                    retVal = compareSpecies( _a );
                    break;
                case Grouping.ORDER_PERSON:
                    retVal = compareAddress( _a );
                    break;
            }
        
        // still equal? then apply date-time-sorting to in-group anymails
        //
        if ( retVal == 0
             && Grouping.BY_STATE != Grouping.BY_TIME
             && Grouping.ORDER_STATE != Grouping.ORDER_TIME
            )
            retVal = compareTime( _a );
        
        return retVal;
    }
    
    int compareSpecies ( Anymail _a )
    {
        return type < _a.type ? -1 : ( type == _a.type ? 0 : 1 );   
    }
    
    int compareTime ( Anymail _a )
    {
        return -((new Date(email.date_received*(long)1000)).compareTo(new Date(_a.email.date_received*(long)1000)));
        
        /*if ( retVal == 0 )
            retVal = email.date_received < _a.email.date_received ? -1
                     : ( email.date_received == _a.email.date_received ? 0 : 1 );*/
    }
    
    int compareAge ( Anymail _a )
    {
        return ( age < _a.age ? 1 : (age == _a.age ? 0 : -1) );
    }
    
    int compareStatus ( Anymail _a )
    {
        int i1 = !email.read    ? 0 : (!email.answered    ? 1 : 2);
        int i2 = !_a.email.read ? 0 : (!_a.email.answered ? 1 : 2);
        return ( i1 > i2 ? 1 : ( i1==i2 ? 0 : -1 ) );
    }
    
    int compareAddress ( Anymail _a )
    {
        String a1 = applemailreader.addresses[applemailreader.addressMap.indexForValue(email.sender_id)].address;
        String a2 = applemailreader.addresses[applemailreader.addressMap.indexForValue(_a.email.sender_id)].address;
        return a1.compareTo( a2 ) < 0 ? -1 : (a1.equals(a2) ? 0 : 1);
    }
}



 /*    Species
  *
  *    species is a class to hold everything needed for drawing (sprites)
  *    and animating (movements, rotations).
  *    i introduced this class mainly to allow for many anymails to share
  *    certain common elements and ease memory usage.
  */

// these hold an index into the species array:
// species[TYPE_JOB] for example.

static int TYPE_FAMILYFRIENDS;
static int TYPE_JOB;
static int TYPE_SCHOOL;
static int TYPE_ECOMMERCE;
static int TYPE_SPAM;
static int TYPE_UNCLASSIFIED;


class Species
{
         /* [state][age] */
    FloatAnimator[][]  movements;
    FloatAnimator[][]  rotations;
    SpriteAnimator[][] sprites;
    
    int state = 0;
    int age = 0;
       
    final static int ROTATION = 0;
    final static int MOVEMENT = 1;
    final static int SPRITE   = 2;
    
    Species ( SpriteContainer[][] _sprites, FloatContainer[][] _movements, FloatContainer[][] _rotations  )
    {
        load( _sprites, _movements, _rotations );
    }
    
    Species ( SpriteAnimator[][] _sprites, FloatAnimator[][] _movements, FloatAnimator[][] _rotations  )
    {
        sprites = new SpriteAnimator[_sprites.length][0];
        movements = new FloatAnimator[_movements.length][0];
        rotations = new FloatAnimator[_rotations.length][0];
        
        for ( int i = 0; i < _sprites.length; i++ )
        {
            sprites[i] = new SpriteAnimator[_sprites[i].length];
            movements[i] = new FloatAnimator[_movements[i].length];
            rotations[i] = new FloatAnimator[_rotations[i].length];
            
            for ( int ii = 0; ii < _sprites[i].length; ii++ )
            {
                sprites[i][ii] = new SpriteAnimator( _sprites[i][ii].container );
                movements[i][ii] = new FloatAnimator( _movements[i][ii].container );
                rotations[i][ii] = new FloatAnimator( _rotations[i][ii].container );
            }
        }
    }
    
    void load ( SpriteContainer[][] _sprites, FloatContainer[][] _movements, FloatContainer[][] _rotations  )
    {
        sprites = new SpriteAnimator[_sprites.length][0];
        movements = new FloatAnimator[_movements.length][0];
        rotations = new FloatAnimator[_rotations.length][0];
        
        for ( int i = 0; i < _sprites.length; i++ )
        {
            sprites[i] = new SpriteAnimator[_sprites[i].length];
            movements[i] = new FloatAnimator[_movements[i].length];
            rotations[i] = new FloatAnimator[_rotations[i].length];
            
            for ( int ii = 0; ii < _sprites[i].length; ii++ )
            {
                sprites[i][ii] = new SpriteAnimator( _sprites[i][ii] );
                movements[i][ii] = new FloatAnimator( _movements[i][ii] );
                rotations[i][ii] = new FloatAnimator( _rotations[i][ii] );
            }
        }
    }
    
    void setState ( int _state )
    {
        state = min( _state, sprites.length-1 );
    }
    
    void setAge ( float _val ) // 0.0 -> 1.0 where 0.0 = oldest, 1.0 is youngest
    {
        //if ( _val >= 0.0 && _val <= 1.0 )
            age = (int)( (sprites[state].length-1) * (1.0 - _val) );
    }
    
    void setAge ( int _age )
    {
        age = min( _age, sprites[state].length-1 );
    }
    
    float rotation () { return rotation( state, age ); }
    float rotation ( int _state, int _age )
    {
        return rotations[_state][_age].value();
    }
    
    float movement () { return movement( state, age ); }
    float movement ( int _state, int _age )
    {
        return movements[_state][_age].value();
    }
    
    PImage sprite () { return sprite( state, age ); }
    PImage sprite ( int _state, int _age )
    {
        return sprites[_state][_age].image();
    }
    
    void tick () { tick( state, age, 1 ); }
    void tick ( int _state, int _age, int _a )
    {
        sprites[_state][_age].tick(_a);
        rotations[_state][_age].tick(_a);
        movements[_state][_age].tick(_a);
    }
    
    Species clone ()
    {
        return new Species(sprites,movements,rotations);
    }
}


  /*    Area
   *
   *    area is a container for anymails, they are confined to this space.
   *    besides that it's used to clear the screen every frame and
   *    implements a mousemanager to facilitate roll-overs and such.
   */

class Area
extends MausManager
{
    int x, y, width, height;
    int xwidth, yheight;
    int xwidthHalf, yheightHalf;
    int padding;
    int xpad, ypad, wpad, hpad;
 
    Area ( int _x, int _y, int _w, int _h )
    {
        x = _x;
        y = _y;
        
        width = _w;
        height = _h;
        
        xwidth = x+width;
        xwidthHalf = xwidth / 2;
        
        yheight = y+height;
        yheightHalf = yheight / 2;
        
        padding = 5;
        xpad = x+padding;
        ypad = y+padding;
        wpad = xwidth-padding;
        hpad = yheight-padding;
    }
    
    boolean intersects ( float rx, float ry, float rw, float rh )
    {
        return intersects((int)rx, (int)ry, (int)rw, int(rh));
    }
    
    boolean intersects ( int rx, int ry, int rw, int rh )
    {
        return !( xpad > rx+rw || wpad < rx || ypad > ry+rh || hpad < ry );
    }
    
    boolean inside ( float rx, float ry, float rw, float rh )
    {
        return inside((int)rx, (int)ry, (int)rw, int(rh));
    }
    
    boolean inside ( int rx, int ry, int rw, int rh )
    {
        return !( xpad > rx || wpad < rx+rw || ypad > ry || hpad < ry+rh );
    }
    
    void draw ( int _color )
    {
        fill( _color );
        rect(x,y,width,height);   
    }
}


  /*  AnymailComperator
   *  
   *  used for grouping the Anymails ...
   *  see compareTo in class Anymail.
   *
   *  takes two Emlx (mails) and passes the comparison on 
   *  to the Anymail containing it.
   */
 
class AnymailComperator
implements Comparator
{
    public int compare ( Object o1, Object o2 ) 
    {
        //if( o1 instanceof Emlx && o2 instanceof Emlx ) { // for speed we just assume this is true
            Anymail a1 = anymails[anymailToEmailId.indexForValue(((Emlx)o1).id)];
            Anymail a2 = anymails[anymailToEmailId.indexForValue(((Emlx)o2).id)];
                        
            return a1.compareTo(a2);
        //}
        //return 0;
    }
}


class GridLookUp
{
    int gridSize = 6; // this is not in pixels, but pixels = pow(2,gridSize)
                      // 1 = 2px, 2 = 4px, 3 = 8px, 4 = 16px, ...
                      // real value will be calculated and kept in gridSizeP.
    int gridSizeP;
                      
    Anymail[][][] table;  // [x][y][]
    
    int wGrid, hGrid;
    
    GridLookUp ( int _w, int _h, int _g )
    {
        gridSize = _g;
        gridSizeP = (int)fastPow(2,gridSize);
        wGrid = (_w + gridSizeP) >> gridSize;  // ceil!
        hGrid = (_h + gridSizeP) >> gridSize;
    }
    
    void clear ()
    {
        table = new Anymail[wGrid][hGrid][0];
    }
    
    int keyForValue ( float _val )
    {
        return keyForValue((int)_val);
    }
    
    int keyForValue ( int _val )
    {
        return _val >> gridSize;  // floor!
    }
    
    void add ( Anymail _a )
    {
        int xKey = keyForValue(_a.loc.x); _a.xHashKey = xKey;
        int yKey = keyForValue(_a.loc.y); _a.yHashKey = yKey;
        
        if ( table[xKey][yKey] != null && table[xKey][yKey].length > 0  )
        {
            Anymail[] tmp = new Anymail[ table[xKey][yKey].length+1 ];
            System.arraycopy( table[xKey][yKey], 0, tmp, 0, table[xKey][yKey].length );
            tmp[table[xKey][yKey].length] = _a;
            table[xKey][yKey] = tmp;
        }
        else
        {
            table[xKey][yKey] = new Anymail[]{_a};
        }
    }
    
    Anymail[] get ( Anymail _a )
    {
        int xKey = _a.xHashKey; int xk = xKey-1;
        int yKey = _a.yHashKey; int yk = yKey-1;
        
        Anymail[] neighbors = new Anymail[0];
        
        // top-left
        if ( xk >= 0 && yk >= 0 && table[xk][yk] != null && table[xk][yk].length > 0 )
        {
            neighbors = concatAnimails(neighbors,table[xk][yk]);
        }
        xk++;
        // top
        if ( yk >= 0 && table[xk][yk] != null && table[xk][yk].length > 0 )
        {
            neighbors = concatAnimails(neighbors,table[xk][yk]);
        }
        xk++;
        // top-right
        if ( xk < wGrid && yk >= 0 && table[xk][yk] != null && table[xk][yk].length > 0 )
        {
            neighbors = concatAnimails(neighbors,table[xk][yk]);
        }
        xk=xKey-1; yk++;
        // left
        if ( xk >= 0 && table[xk][yk] != null && table[xk][yk].length > 0 )
        {
            neighbors = concatAnimails(neighbors,table[xk][yk]);
        }
        xk++;
        // center
        if ( table[xk][yk] != null && table[xk][yk].length > 0 )
        {
            neighbors = concatAnimails(neighbors,table[xk][yk]);
        }
        xk++;
        // right
        if ( xk < wGrid && table[xk][yk] != null && table[xk][yk].length > 0 )
        {
            neighbors = concatAnimails(neighbors,table[xk][yk]);
        }
        xk=xKey-1; yk++;
        // bottom-left
        if ( xk >= 0 && yk < hGrid && table[xk][yk] != null && table[xk][yk].length > 0 )
        {
            neighbors = concatAnimails(neighbors,table[xk][yk]);
        }
        xk++;
        // bottom-left
        if ( yk < hGrid && table[xk][yk] != null && table[xk][yk].length > 0 )
        {
            neighbors = concatAnimails(neighbors,table[xk][yk]);
        }
        xk++;
        // bottom-left
        if ( xk < wGrid && yk < hGrid && table[xk][yk] != null && table[xk][yk].length > 0 )
        {
            neighbors = concatAnimails(neighbors,table[xk][yk]);
        }
        
        return neighbors;
    }
    
    public String toString()
    {
        String tab = "\t";
        return gridSize + tab + gridSizeP + tab + wGrid + tab + hGrid;
    }
}

final static Anymail[] concatAnimails ( Anymail[] a1, Anymail[] a2 )
{
    //if ( a1 == null || a2 == null ) return null; // to gain speed we assume this is always true
    
    Anymail[] tmp = new Anymail[a1.length+a2.length];
    System.arraycopy( a1, 0, tmp, 0, a1.length );
    System.arraycopy( a2, 0, tmp, a1.length, a2.length );
    return tmp;
}



 
/*
 *    This work is licensed under a 
 *    Creative Commons Attribution-Noncommercial-Share Alike 3.0 License.
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/
 */
