
/*    in here are the basic classes for the interface
 *    elements. these are extended by the actual 
 *    implementations (see tabs InterfaceExtended, InterfaceExtended2).
 */
 
 
/*    InterfaceElement
 *    
 *    the most basic interface class. it implements 
 *    MausDing, which is part of the maus (mouse) library.
 *    (see http://bezier.de/maus/)
 */

interface InterfaceElement
{
    void draw ();
    void draw ( boolean _force );
    void drawSelf ();
}

class InterfaceElementBasic
implements MausDing, InterfaceElement
{
    Object[] listeners;
    
    String label;
    int x,y,w,h;
    
    boolean dirty = true;
    
    long callbackDelay = 0;
    long lastCallback;
    
    int fColor = Colors.VIVID_MEDIUM_GREEN;
    int lColor = Colors.DARK_MEDIUM_GREEN;
    
    InterfaceElementBasic () {}
    
    InterfaceElementBasic ( int _x, int _y, int _w, int _h, String _label ) {
        x = _x; y = _y; w = _w; h = _h; label = _label;
    }
    
    void draw ( boolean _force ) { dirty = _force; draw(); }
    
    void draw () {
        if ( !dirty ) return;
        drawSelf();
        dirty = false;
    }
    
    void drawSelf () {}
    
    
    // ---- interface MausDing -----
    
    boolean check (int _mx, int _my) {
        if ( _mx < x || _mx > x+w || _my < y || _my > y + h ) return false;
        return true;
    }
	
    void enter () { dirty = true; }

    void over () { dirty = true; /*if ((debug & INTERFACE) == INTERFACE) println(this.getClass().getName());*/ }

    void exit () { dirty = true; }

    void pressedRight () { dirty = true; }

    void releasedOutsideRight () { dirty = true; }

    void releasedRight () { dirty = true; }

    void draggedRight ( int _dx, int _dy ) { dirty = true; }

    void pressed () { dirty = true; }

    void releasedOutside () { dirty = true; }

    void released () { dirty = true; }
	
    void dragged (int _dx, int _dy) { dirty = true; }
    
    
    // ---- callbacks & listeners -----
    
    void registerListener ( Object _o  )
    {
        if ( listeners == null ) listeners = new Object[]{_o};
        else
        {
            Object[] _objs = new Object[listeners.length+1];
            System.arraycopy(listeners, 0, _objs, 0, listeners.length);
            _objs[_objs.length-1] = _o;
            listeners = _objs;
        }
    }
    
    void callListeners ( String _meth ) { callListeners( _meth, new Object[]{} ); }
    void callListeners ( String _meth, Object _o ) { callListeners( _meth, new Object[]{_o} ); }
    void callListeners ( String _meth, Object[] _args )
	{
		if ( listeners == null || listeners.length == 0 ) return;
		
                if ( lastCallback + callbackDelay > System.currentTimeMillis() ) return;

		int i = listeners.length-1;
		
		Class[] cs = new Class[_args.length];
		for ( int k = 0; k < _args.length; k++ )
		{
			cs[k] = _args[k].getClass();
		}
		
		do { try {
                                if ( listeners[i] == null ) { 
                                    println(getClass().getName() + " /ERROR/ Listener pale: "+i); 
                                    i--; continue;
                                }
				Class c = listeners[i].getClass();
				Method m = c.getMethod( _meth, cs );
				m.invoke( listeners[i], _args );
			} catch ( InvocationTargetException ite  ) {//  ite.printStackTrace();
			} catch ( NoSuchMethodException     nsme ) {// nsme.printStackTrace();
			} catch ( IllegalArgumentException  iare ) {// iare.printStackTrace();
			} catch ( IllegalAccessException    iace ) {// iace.printStackTrace();
			} 
                        i--;
		} while ( i >= 0 );
            
                lastCallback = System.currentTimeMillis();
	}
}


/*    InterfaceContainer
 *    
 *    a basic class for elements that house sub-elements.
 */

class InterfaceContainer
extends MausManager
implements InterfaceElement, MausDing
{
    Object[] listeners;
    
    InterfaceElement[] elements;
    
    String label;
    int x,y,w,h;
    
    boolean dirty = true;
    boolean enabled = true; // need to disable on creation if it's supposed to be hidden
    
    int fColor = Colors.VIVID_MEDIUM_GREEN;
    int lColor = Colors.DARK_MEDIUM_GREEN;
    
    InterfaceContainer () {}
    
    InterfaceContainer ( int _x, int _y, int _w, int _h, String _label ) {
        x = _x; y = _y; w = _w; h = _h; label = _label;
    }
    
    void draw ( boolean _force ) { 
        if ( !enabled ) return;
        dirty = _force; draw(); 
    }
    
    void clear () { }
    
    void draw ()
    {
        if ( !enabled ) return;
        if ( dirty ) drawSelf();
        drawElements();
        dirty = false;
    }
    
    void drawSelf () {}
    
    void drawElements ( )
    {
        if ( !enabled || elements == null ) return;
        
        int i=0; try { do {
        
            elements[i].draw( dirty );
        
        i++; } while(true); } catch (ArrayIndexOutOfBoundsException aioobe) { }
    }
    
    void enable () {
        activated = true;
        dirty = true;
        enabled = true;
    }
    
    void disable () {
        activated = false;
        if ( enabled ) clear();
        dirty = false;
        enabled = false;
    }
    
    
    // ---- interface MausDing -----
    
    boolean check (int _mx, int _my) {
        if ( _mx < x || _mx > x+w || _my < y || _my > y + h ) return false;
        return true;
    }
	
    void enter () { dirty = true; }

    void over () { dirty = true; }

    void exit () { dirty = true; } // will be cleared with the next check

    void pressedRight () { dirty = true;}

    void releasedOutsideRight () { dirty = true; } // will be cleared with the next check

    void releasedRight () { dirty = true; }

    void draggedRight ( int _dx, int _dy ) { dirty = true; }

    void pressed () { dirty = true; }

    void releasedOutside () { dirty = true; } // will be cleared with the next check

    void released () { dirty = true; }
	
    void dragged (int _dx, int _dy) { dirty = true; }
    
    
    // ---- callbacks & listeners -----
    
    void registerListener ( Object _o  )
    {
        if ( listeners == null ) listeners = new Object[]{_o};
        else
        {
            Object[] _objs = new Object[listeners.length+1];
            System.arraycopy(listeners, 0, _objs, 0, listeners.length);
            _objs[_objs.length-1] = _o;
            listeners = _objs;
        }
    }
    
    void callListeners ( String _meth ) { callListeners( _meth, new Object[]{} ); }
    void callListeners ( String _meth, Object _o ) { callListeners( _meth, new Object[]{_o} ); }
    void callListeners ( String _meth, Object[] _args )
	{
		if ( listeners == null || listeners.length == 0 ) return;
		
		int i = listeners.length-1;
		
		Class[] cs = new Class[_args.length];
		for ( int k = 0; k < _args.length; k++ )
		{
			cs[k] = _args[k].getClass();
		}
		
		do { try {
                                if ( listeners[i] == null ) { 
                                    println(getClass().getName() + " /ERROR/ Listener pale: "+i); 
                                    i--; continue;
                                }
				Class c = listeners[i].getClass();
				Method m = c.getMethod( _meth, cs );
				m.invoke( listeners[i], _args );
			
				i--;
			} catch ( InvocationTargetException ite  ) {  ite.printStackTrace(); i--;
			} catch ( NoSuchMethodException     nsme ) { nsme.printStackTrace(); i--;
			} catch ( IllegalArgumentException  iare ) { iare.printStackTrace(); i--;
			} catch ( IllegalAccessException    iace ) { iace.printStackTrace(); i--;
			}
		} while ( i >= 0 );
	}
}



 
/*
 *    This work is licensed under a 
 *    Creative Commons Attribution-Noncommercial-Share Alike 3.0 License.
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/
 */
