
/*    ... contains some classes that extend InterfaceElement.
 *    these are the actual implementations, see InterfaceBasics
 *    for the prototypes.
 *
 *    DateController
 *    Button
 *    ViewSlider
 *    ViewSliderHandle
 *    DateSlider
 *    ImageButton
 */


class InterfaceContainerImpl
extends InterfaceContainer
{
    InterfaceContainerImpl ( int _x, int _y, int _w, int _h, String _label ) {
        super( _x, _y, _w, _h, _label );
        fColor = Colors.VIVID_GREEN;
    }

    void over () {
        super.over();
        lColor = Colors.DARK_GREEN;
    }

    void exit () {
        super.exit();
        lColor = Colors.DARK_MEDIUM_GREEN;
    }

    void drawSelf ()
    {
        fill( fColor );
        noStroke();
        //if ( (debug & INTERFACE) == INTERFACE ) stroke( lColor );
        //rect(x,y,w,h);
    }
    
    void clear ()
    {
        fill( fColor );
        noStroke();
        rect(x,y,w,h);
    }
}


interface ButtonListener
{
    void buttonReleased ();
}

class Button
extends InterfaceElementBasic
{
    int tColor;
    boolean gotPressed = false; // this enables a continuous pressed state for the slider
    boolean catchContinuous = false;
    boolean selected = false;

    Button ( int _x, int _y, int _w, int _h, String _l ) {
        super( _x, _y, _w, _h, _l );
        tColor = Colors.VIVID_MEDIUM_GREEN;
        fColor = Colors.DARK_GREEN;
        callbackDelay = 200;     // every 200th millisecond,
        // makes the slider less jumpy.
    }

    void pressed () {
        super.pressed();
        gotPressed = true;
    }

    void dragged ( int _dx, int _dy ) {
        super.dragged( _dx, _dy );
        if ( catchContinuous )
            callListeners( "buttonReleased" );
    }

    void released () {
        super.released();
        if ( (catchContinuous && !gotPressed) || !catchContinuous )
        {
            callListeners( "buttonReleased" );
            selected = true;
        }
        
        tColor = Colors.VIVID_MEDIUM_GREEN;
        lColor = Colors.VIVID_MEDIUM_GREEN;
            
        gotPressed = false;
    }

    void over () {
        super.over();
        tColor = Colors.WHITE;
        lColor = Colors.WHITE;
    }

    void exit () {
        super.exit();
        tColor = Colors.VIVID_MEDIUM_GREEN;
        lColor = Colors.VIVID_MEDIUM_GREEN;
    }
    
    void deselect ()
    {
        selected = false;
        dirty = true;
    }

    void drawSelf ()
    {
        fill( fColor );
        noStroke();
        //stroke( lColor );
        rect(x,y,w,h);
        
        noStroke();
        fill( selected ? Colors.WHITE : tColor );
        textAlign( LEFT );
        textFont( BOLD_10 );
        text( label, x, y+8 );
    }
}

interface ViewSliderHandleListener
{
    void viewSliderHandleReleased ( Integer _step );
    void viewSliderHandleChanged  ( Integer _step );
}

class ViewSlider
extends InterfaceContainer
{    
    ViewSlider ( int _x, int _y, int _w, int _h, String _label )
    {
        super( _x, _y, _w, _h, _label );
        fColor = Colors.VIVID_GREEN;
    }

    void drawSelf ()
    {
        fill( fColor );
        noStroke();
        //stroke( lColor );
        rect(x,y,w,h);
    }
}

class ViewSliderHandle
extends InterfaceElementBasic
implements ViewSliderHandleListener
{
    int xBox;
    int yBox;
    int limitLeft;
    int limitRight;
    int hColor;
    PImage handleNormal;
    PImage handleOver;
    PImage handle;

    int step;
    int type; // 0 = left, 1 = right handle

        ViewSliderHandle ( int _x, int _y, int _w, int _h, String _label, 
    int _ll, int _lr, PImage _hn, PImage _ho, int _xs, int _t )
    {
        super( _x, _y, _w, _h, _label );
        xBox = _x;
        yBox = _y;
        step = _xs;
        x += _xs * 20;
        limitLeft = _ll+40;
        limitRight = _lr-40;
        handleNormal = _hn;
        handleOver = _ho;
        handle = handleNormal;
        type = _t;
    }

    void over () {
        super.over();
        handle = handleOver;
        lColor = Colors.WHITE;
        fColor = Colors.VIVID_GREEN;
    }

    int xPrevious;

    void pressed () {
        super.pressed();
        xPrevious = x;
        handle = handleOver;
        lColor = Colors.WHITE;
        fColor = Colors.VIVID_GREEN;
    }

    void dragged ( int _dx, int _dy )
    {
        handle = handleOver;
        int xNext = xPrevious + floor(_dx / 20) * 20;

        if ( xNext >= xBox + limitLeft && xNext <=  xBox + limitRight ) {
            x = xNext;
            step = floor( (x-xBox) / 20);
            callListeners( "viewSliderHandleChanged", new Integer(step) );
        }
    }

    void exit () {
        super.exit();
        handle = handleNormal;
        lColor = Colors.VIVID_MEDIUM_GREEN;
        fColor = Colors.VIVID_GREEN;
    }

    void released () {
        super.released();
        callListeners( "viewSliderHandleReleased", new Integer(step) );
    }

    void releasedOutside () {
        super.releasedOutside();
        callListeners( "viewSliderHandleReleased", new Integer(step) );
    }

    void drawSelf ()
    {
        fill( fColor );
        stroke( Colors.DARK_GREEN );
        //rect( xBox+limitLeft, y, limitRight, h );
        //rect( x, y, w, h );
        fill( hColor );
        noStroke();
        image( handle, x+2, y+2 );
    }

    void viewSliderHandleReleased ( Integer _step ) {
        ;
    }

    void viewSliderHandleChanged( Integer _step )
    {
        int st = _step.intValue();
        if ( st == step ) return;
        if ( type == 1 )
            limitLeft = (st+1)*20;
        else
            limitRight = (st-1)*20;
    }
}

class ImageButton
extends Button
{
    PImage imageNormal, imageOver, imageC;

    ImageButton ( int _x, int _y, int _w, int _h, String _label, PImage _in, PImage _io )
    {
        super( _x, _y, _w, _h, _label );
        fColor = Colors.VIVID_GREEN;
        imageNormal = _in;
        imageOver = _io;
        imageC = imageNormal;
        callbackDelay = 50;     // every 200th millisecond,
        // makes the slider less jumpy.
    }
    
    void enter () {
        super.enter();
        imageC = imageOver;   
    }
    
    void pressed () {
        super.pressed();
        imageC = imageOver;   
    }
    
    void over () {
        super.over();
        imageC = imageOver;   
    }
    
    void exit () {
        super.exit();
        imageC = imageNormal;   
    }
    
    void releasedOutside () {
        super.releasedOutside();
        imageC = imageNormal;   
    }

    void drawSelf ()
    {
        fill( fColor );
        noStroke();
        //stroke( lColor );
        rect(x,y,w,h);
        image( imageC, x,y );
    }
}

class ImageCheckbox
extends ImageButton
{
    int xi, yi;
    boolean checked = false;
    
    ImageCheckbox  ( int _x, int _y, int _w, int _h, String _label, PImage _in, PImage _io )
    {
        super( _x, _y, _w, _h, _label, _in, _io );
        fColor = Colors.BEIGE_MEDIUM;
        imageNormal = _in;
        imageOver = _io;
        imageC = imageNormal;
        xi = x + (w / 2) - (imageNormal.width  / 2);
        yi = y + (h / 2) - (imageNormal.height / 2);
    }
    
    void enter () 
    {
        super.enter();
        fColor = Colors.WHITE;
    }
    
    void over ()
    {
        super.over();
        fColor = Colors.WHITE;
        callListeners( "setLabel", label );
    }
    
    void exit ()
    {
        super.over();
        fColor = Colors.BEIGE_MEDIUM;
        callListeners( "resetLabel" );
    }
    
    void released ()
    {
        super.released();
        checked = !checked;
    }

    void drawSelf ()
    {
        fill( checked ? Colors.VIVID_LIGHT_GREEN : fColor );
        noStroke();
        //stroke( lColor );
        rect(x+1,y,w-1,h);
        image( checked ? imageOver : imageNormal, xi,yi );
    }
}

class TextLabel
extends InterfaceElementBasic
{
    String originalLabel;
    
    TextLabel ( int _x, int _y, int _w, int _h, String _l )
    {
        super( _x, _y, _w, _h, _l );
        originalLabel = _l;
    }
    
    void resetLabel ()
    {
        label = originalLabel;   
    }
    
    void setLabel (String _l)
    {
        label = _l;   
    }

    void drawSelf ()
    {
        fill( Colors.VIVID_GREEN );
        noStroke();
        //stroke( lColor );
        rect(x-1,y-h+1,w+2,h+2);
        fill( Colors.WHITE );
        textAlign( LEFT );
        textFont( BOLD_9 );
        text( label, x+2, y );
    }
}

class CalendarSlider
extends InterfaceElementBasic
{
    ViewSliderHandle leftHandle, rightHandle;
    
    long min_date, max_date;
    long min_hist_date, max_hist_date;
    long min_sel_date, max_sel_date;
    long min_curr_date, max_curr_date;

    long display_month;
    long display_next_month;

    int  display_w = 21;
    int  display_ws = 20;

    int  display_days = 44;
    int  display_first_day = 0;
    int  total_days;
    
    long timezone_offset;

    long days_in_display;

    long date_selection_start, date_selection_end;

    java.text.SimpleDateFormat dateformat;

    CalendarSlider ( int _x, int _y, int _w, int _h, String _l, long _mid, long _mad )
    {
        super( _x, _y, _w, _h, _l );

        timezone_offset = Calendar.getInstance().getTimeZone().getRawOffset();

        min_date = floor( _mid / Timeperiod.DAY ) * Timeperiod.DAY - Timeperiod.DAY * 2 - Timeperiod.HOUR - timezone_offset;
        max_date =  ceil( _mad / Timeperiod.DAY ) * Timeperiod.DAY + Timeperiod.DAY * 2 - timezone_offset;

        min_curr_date = max_date - ((long)(display_days-1) * Timeperiod.DAY);

        date_selection_start = min_curr_date + (long)4 * Timeperiod.DAY;
        date_selection_end   = min_curr_date + (long)10 * Timeperiod.DAY;

        total_days = ceil( (max_date - min_date) / Timeperiod.DAY );

        days_in_display = (long)display_days * Timeperiod.DAY;

        dateformat = new SimpleDateFormat("MMM  yyyy");
    }
    
    int rollDist = 0;
    boolean insideSelection = false;
    int xStart, xEnd;
    
    void over ()
    {
        super.over();
        
        int xClick = int((mouseX - x) / 20.);
        xStart = int((date_selection_start - min_curr_date ) / Timeperiod.DAY);
        xEnd   = int((date_selection_end   - min_curr_date ) / Timeperiod.DAY);
        
        insideSelection = xClick >= xStart && xClick < xEnd;
        if ( insideSelection )
        {
            leftHandle.over();
            rightHandle.over();
        }
    }
    
    void pressed ()
    {
        super.pressed();
        
        rollDist = 0;
        
        if ( insideSelection )
        {
            leftHandle.pressed();
            rightHandle.pressed();
        }
    }
    
    void dragged ( int _dx, int _dy )
    {
        super.dragged( _dx, _dy );
        int rDist = -1 * floor(_dx / 21.);
        if ( rollDist != rDist )
        {
            if ( insideSelection )
            {
                if ( _dx > 0 )
                {
                    if ( xEnd - rDist < display_days - 1 )
                        rightHandle.dragged(_dx,0);
                    
                    if ( xStart - rDist > 1 )
                        leftHandle.dragged(_dx,0);
                }
                else
                {
                    if ( xStart - rDist > 1 )
                        leftHandle.dragged(_dx,0);
                        
                    if ( xEnd - rDist < display_days - 1 )
                        rightHandle.dragged(_dx,0);
                }
            }
            else
            {
                roll( rDist - rollDist );
            }
            
            rollDist = rDist;
        }
    }
    
    void releasedOutside ()
    {
        super.releasedOutside();
        released();
    }
    
    void released ()
    {
        super.released();
        if ( insideSelection )
        {
            if ( rollDist > 0 )
            {
                rightHandle.exit();
                leftHandle.exit();
            }
            else
            {
                leftHandle.exit();
                rightHandle.exit();
            }
        }
        insideSelection = false;   
    }
    
    void exit ()
    {
        super.exit();
        
        rightHandle.exit();
        leftHandle.exit();   
    }

    void drawSelf ()
    {
        fill( fColor );
        noStroke();
        //stroke( lColor );
        //rect(x,y,w,h);
        stroke( Colors.VIVID_GREEN );
        textAlign( CENTER );
        fColor = Colors.VIVID_LIGHT_GREEN;
        textFont( BOLD_9 );
        
        long dateits;
        
        int[] lTextLast = new int[]{0,0};
        int[] rTextLast = new int[]{0,0};

        for ( int i = 0; i < display_days; i++ )
        {
            int xd = x + i*display_ws;

            Calendar cal = Calendar.getInstance(); // relative to local time

                cal.setTimeInMillis( min_curr_date );
            cal.add( Calendar.DAY_OF_YEAR, i );
            dateits = cal.getTimeInMillis();

            if ( cal.get( Calendar.MONTH ) % 2 == 0 ) // month even or uneven?
            {
                fColor = Colors.VIVID_LIGHT_GREEN;
            }
            else
            {
                fColor = Colors.VIVID_MEDIUM_GREEN;
            }

            if ( dateits >= date_selection_start && dateits < date_selection_end )
            {
                fColor = Colors.WHITE;
            }

            fill( fColor );
            rect( xd, y, display_w, h );

            fill( Colors.GREY_DARK );
            text( cal.get( Calendar.DAY_OF_MONTH ), xd+10, y+12 );

            if ( dateits - date_selection_start == 0 ) // start selection part ..
            {
                fill(Colors.WHITE);
                textAlign(RIGHT);
                text( dateformat.format(new Date(cal.getTimeInMillis())).toUpperCase(), xd-9, y+h+6+7);

                textAlign( CENTER );
            }
            if ( dateits - date_selection_end == 0 ) // end selection part ..
            {
                fill(Colors.WHITE);
                textAlign(LEFT);
                text( dateformat.format(new Date(cal.getTimeInMillis())).toUpperCase(), xd+9, y+h+6+7);

                textAlign( CENTER );
            }
        }
    }

    void roll ( int days )
    {
        Calendar cal = Calendar.getInstance();
        cal.setTimeInMillis( min_curr_date );
        cal.add( Calendar.DAY_OF_MONTH, days );

        long to_date = cal.getTimeInMillis();

        if ( to_date >= min_date && to_date <= max_date - days_in_display + Timeperiod.DAY )
        {
            min_curr_date = to_date;
            
            cal.setTimeInMillis( date_selection_start );
            cal.add( Calendar.DAY_OF_MONTH, days );
            date_selection_start = cal.getTimeInMillis();
            
            cal.setTimeInMillis( date_selection_end );
            cal.add( Calendar.DAY_OF_MONTH, days );
            date_selection_end = cal.getTimeInMillis();
        }
    }

    void rollMonth ( int months )
    {
        Calendar cal = Calendar.getInstance();
        cal.setTimeInMillis( date_selection_start );
        roll( months * cal.getActualMaximum( Calendar.DAY_OF_MONTH ) ); // pivot start date of selection
    }

    void setSelection ( int dayDist, int handle )
    {
        Calendar cal = Calendar.getInstance();
        cal.setTimeInMillis( min_curr_date );
        cal.add( Calendar.DAY_OF_MONTH, dayDist );

        //println( new Date(cal.getTimeInMillis()) );

        if ( handle == 0 && cal.getTimeInMillis() < date_selection_end ) // left handle
        {
            date_selection_start = cal.getTimeInMillis();
        }
        else if ( cal.getTimeInMillis() > date_selection_start )
        {
            date_selection_end = cal.getTimeInMillis();
        }
    }
}

interface GroupingMenuListener
{
    void stateChanged (Integer _state) ;
}

class GroupingMenu
extends InterfaceElementBasic
{
    PImage imageEnd;
    PImage menuArrowLeft, menuArrowLeftOver, menuArrowLeftC;
    PImage menuArrowRight, menuArrowRightOver, menuArrowRightC;
    int xil, xir;
    
    PImage[] images;
    String[] choices;
    boolean[] enabledChoices;
    int current = 1;
    
    boolean enabled = true;
    
    GroupingMenu ( int _x, int _y, int _w, int _h, String _l )
    {
        super( _x, _y, _w, _h, _l );
        xil = x+41;
        xir = x+41+13;
    }
    
    boolean check (int _mx, int _my) {
        boolean ret = super.check(_mx,_my);
        if ( !enabled ) return false;
        else return ret;
    }
    
    void over ()
    {
        super.over();
        
        if ( _inside(mouseX,mouseY, xil, y, menuArrowLeft.width, menuArrowLeft.height) )
            menuArrowLeftC = menuArrowLeftOver;
        else
            menuArrowLeftC = menuArrowLeft;
            
        if ( _inside(mouseX,mouseY, xir, y, menuArrowRight.width, menuArrowRight.height) )
            menuArrowRightC = menuArrowRightOver;
        else
            menuArrowRightC = menuArrowRight;
    }
    
    void enter ()
    {
        super.enter();
        
        if ( _inside(mouseX,mouseY, xil, y, menuArrowLeft.width, menuArrowLeft.height) )
            menuArrowLeftC = menuArrowLeftOver;
        else
            menuArrowLeftC = menuArrowLeft;
            
        if ( _inside(mouseX,mouseY, xir, y, menuArrowRight.width, menuArrowRight.height) )
            menuArrowRightC = menuArrowRightOver;
        else
            menuArrowRightC = menuArrowRight;
    }
    
    int _pressed = 0; // 0, 1 = left, 2 = right
    
    void pressed ()
    {
        super.pressed();
        
        _pressed = 0;
        
        if ( _inside(mouseX,mouseY, xil, y, menuArrowLeft.width, menuArrowLeft.height) )
        {
            menuArrowLeftC = menuArrowLeftOver;
            _pressed = 1;
        }
        else
            menuArrowLeftC = menuArrowLeft;
            
        if ( _inside(mouseX,mouseY, xir, y, menuArrowRight.width, menuArrowRight.height) )
        {
            menuArrowRightC = menuArrowRightOver;
            _pressed = 2;
        }
        else
            menuArrowRightC = menuArrowRight;
    }
    
    void released ()
    {
        super.released();
        
        if ( _pressed == 1 && _inside(mouseX,mouseY, xil, y, menuArrowLeft.width, menuArrowLeft.height) )
        {
            menuArrowLeftC = menuArrowLeftOver;
            choose(-1);
            callListeners("stateChanged", new Integer(current));
        }
        else
            menuArrowLeftC = menuArrowLeft;
            
        if ( _pressed == 2 && _inside(mouseX,mouseY, xir, y, menuArrowRight.width, menuArrowRight.height) )
        {
            menuArrowRightC = menuArrowRightOver;
            choose(+1);
            callListeners("stateChanged", new Integer(current));
        }
        else
            menuArrowRightC = menuArrowRight;
            
        _pressed = 0;
    }
    
    void exit ()
    {
        super.exit();
        
        menuArrowLeftC = menuArrowLeft;
        menuArrowRightC = menuArrowRight;
    }
    
    void choose ( int _i )
    {
        current = current + _i;
        current = current < 0 ? choices.length-1 : ( current >= choices.length ? 0 : current  );
        
        while ( enabledChoices != null && !enabledChoices[current] )
        {
            current = current + _i;
            current = current < 0 ? choices.length-1 : ( current >= choices.length ? 0 : current  );
        }
    }
    
    void setEnabled ( boolean[] _ec )
    {
        enabledChoices = _ec;
        if ( !enabledChoices[current] )
        {
            choose(1);
            callListeners("stateChanged", new Integer(current));
        }
    }
    
    boolean _inside ( int _x, int _y, int _xd, int _yd, int _wd, int _hd )
    {
        return !( _x > _xd+_wd || _x < _xd || _y > _yd+_hd || _y < _yd );
    }
    
    void drawSelf ()
    {
        fill( Colors.VIVID_LIGHT_GREEN );
        noStroke();
        rect( x+67 ,y,124,h );
        
        fill( Colors.WHITE );
        rect( x+1,y, 39, h );
        image(images[current], x+20-(images[current].width/2), y+8-(images[current].height/2));
        
        if (imageEnd != null) image( imageEnd, x+w-imageEnd.width,y );
        
        image( menuArrowLeftC,  xil, y );
        image( menuArrowRightC, xir, y );
        
        textFont( BOLD_9 );
        fill( Colors.GREY_DARK );
        text( choices[current], x+41+13+13+1, y+h-5 );
        
        if ( !enabled )
        {
            fill( 0xBB000000 + (Colors.VIVID_GREEN & 0x00FFFFFF) );
            rect( x,y,w,h+20 );
        }
    }
}



 
/*
 *    This work is licensed under a 
 *    Creative Commons Attribution-Noncommercial-Share Alike 3.0 License.
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/
 */
