
/*
 *    All non-menu interface elements:
 *
 *    - ShadowedBox
 *
 */

class ShadowedBox
extends InterfaceContainer
{
    PImage stencil;
    PImage _img;
    
    String title = "";     // BOLD_9, uppercase, white
    String type = "";   // REGULAR_9, white
    String state = "";     // -"-
    String time = "";      // BOLD_9
    String date = "";      // -"-
    String subject = "";   // BOLD_9, GREY_DARK
    
    int color1, color2;
    
    ShadowedBox ( int _x, int _y, int _w, int _h, String _label ) {
        super( _x, _y, _w, _h, _label );
        stencil = loadImage( "_data/interface/shadowedBox/shadowedBoxMask.png" );
        _img = new PImage(_w,_h,ARGB);
    }
    
    ShadowedBox ( int _x, int _y, int _w, int _h, String _label, PImage _stencil ) {
        this( _x, _y, _w, _h, _label );
        stencil = _stencil;
    }
    
    void applyAnymail ( Anymail _a )
    {
        color1 = ANYMAIL_COLORS[_a.type][0];
        color2 = ANYMAIL_COLORS[_a.type][1];
        
        title = _a.name.toUpperCase();
        title = title.equals("") ? applemailreader.addresses[
                                           applemailreader.addressMap.indexForValue(_a.email.sender_id)
                                                           ].address : title;
        type = SPECIES_NAMES[_a.type];
        state = STATES_NAMES[_a.state];
        Date received = new Date( _a.email.date_received * ((long)1000) );
        time = (new SimpleDateFormat("dd:MM:yyyy")).format( received );
        date = (new SimpleDateFormat("HH:mm")).format( received );
        
        subject = "(No subject)";
        if ( _a.email.subject_id > 0 )
        {
            subject = applemailreader.subjects[
                                    applemailreader.subjectMap.indexForValue(_a.email.subject_id)
                                          ].subject;
                                          
            subject = subject.length() > 35 ? subject.substring(0,35)+" ..." : subject;
        }
        dirty = true;
    }
    
    void draw ( int _x, int _y, int _w, int _h )
    {
        if ( _w == 0 ) return;
        if ( _h == 0 ) return;
        if ( _w < stencil.width ) { _w = stencil.width;  }
        if ( _h < stencil.height) { _h = stencil.height; }
        
        x = _x; y = _y; 
        
        if ( dirty || _w != w || _h != h )
        {
            dirty = true;
            _img.pixels = new int[_w*_h];
            _img.width = _w; _img.height = _h;
            _img.updatePixels();
        }
        
        draw ( _x, _y );
        
        dirty = false;
    }

    void draw ( int _x, int _y )
    {
        x = _x; y = _y;
        
        if ( dirty )
        {
            fillRect( 0,  0, _img.width, 17, color1 );
            fillRect( 0, 17, _img.width, 17, color2 );
            fillRect( 0, 34, _img.width, 50, Colors.WHITE );
            
            applyShadow( _img, stencil, 6 );
            
            dirty = false;
        }
        
        drawSelf ();
    }
    
    void drawSelf ()
    {
        image( _img, x, y );
        
        textFont(BOLD_9);
        fill(Colors.WHITE);
        text(title,x+9,y+13);
        
        textFont(REGULAR_9);
        text(type,x+9,y+29);
        
        textAlign(RIGHT);
        text(state,x+_img.width-6-9,y+29);
            
        textAlign(LEFT); 
        textFont(BOLD_9); 
        fill(color1); 
        text(time,x+9,y+51); 
        
        textAlign(RIGHT); 
        text(date,x+_img.width-6-9,y+51);
        
        textAlign(LEFT); 
        fill(Colors.GREY_DARK);
        text(subject,x+9,y+67);
    }
    
    void clear ()
    {
    }

    void applyShadow ( PImage _img, PImage stencil, int grid )
    {
        int grid2 = grid * 2;
        int grid3 = grid * 3;
        
        if ( _img.height < grid3 ) return;
        if ( _img.width  < grid3 ) return;

        _img.loadPixels();

        // top left
        for ( int y = 0; y < grid; y ++ )
            for ( int x = 0; x < grid; x++ )
                _img.pixels[y*_img.width+x] = _applyStencilPixel( _img.pixels[y*_img.width+x], 
                stencil.pixels[y*stencil.width+x] );

        // top right
        int iy = _img.width - grid2;
        int iy2 = stencil.width - grid2;
        for ( int y = 0; y < grid2; y ++ )
            for ( int x = 0; x < grid2; x++ )
                _img.pixels[iy + y*_img.width+x] = _applyStencilPixel( _img.pixels[iy + y*_img.width+x], 
                stencil.pixels[iy2 + y*stencil.width+x] );

        // top right - bottom right
        iy = _img.width - grid2 + grid2 * _img.width;
        iy2 = stencil.width - grid2 + grid2 * stencil.width;
        int hh = _img.height - grid2 - grid3;
        for ( int y = 0; y < hh; y ++ )
            for ( int x = 0; x < grid2; x++ )
                _img.pixels[iy + y*_img.width+x] = _applyStencilPixel( _img.pixels[iy + y*_img.width+x],
                stencil.pixels[iy2 + stencil.width+x] );

        // bottom left
        iy = _img.pixels.length - grid2 * _img.width;
        iy2 = stencil.pixels.length - grid2 * stencil.width;
        for ( int y = 0; y < grid2; y ++ )
            for ( int x = 0; x < grid2; x++ )
                _img.pixels[iy + y*_img.width+x] = _applyStencilPixel( _img.pixels[iy + y*_img.width+x],
                stencil.pixels[iy2 + y*stencil.width+x] );


        // bottom left - bottom right
        iy2 = stencil.pixels.length - grid2 * stencil.width + grid2;
        int ww = _img.width - grid3;
        for ( int y = 0; y < grid2; y ++ )
            for ( int x = grid2; x < ww; x++ )
                _img.pixels[iy + y*_img.width+x] = _applyStencilPixel( _img.pixels[iy + y*_img.width+x],
                stencil.pixels[iy2 + y*stencil.width] );


        // bottom right
        iy = _img.pixels.length - grid3 * _img.width + _img.width - grid3;
        iy2 = stencil.pixels.length - grid3 * stencil.width + grid2;
        for ( int y = 0; y < grid3; y ++ )
            for ( int x = 0; x < grid3; x++ )
                _img.pixels[iy + y*_img.width+x] = _applyStencilPixel( _img.pixels[iy + y*_img.width+x],
                stencil.pixels[iy2 + y*stencil.width+x] );

        _img.updatePixels();
    }

    int _applyStencilPixel ( int t, int f )
    {
        int r = (f & 0xFF000000) | ( (( 0xFF00FFFF | f ) > 0xFF3aFFFF) ? (t & 0x00FFFFFF) : Colors.BOX_SHADOW );
        return r;
    }

    void fillRect ( int x, int y, int w, int h, int col )
    {
        x = x >= _img.width  ? _img.width-1  : ( x < 0 ? 0 : x );
        y = y >= _img.height ? _img.height-1 : ( y < 0 ? 0 : y );
        w = w + x > _img.width  ? _img.width  - x : ( w <= 0 ? 1 : w );
        h = h + y > _img.height ? _img.height - y : ( h <= 0 ? 1 : h );
        
        int[] row = new int[w];
        for (int i = 0; i < w; i++) row[i] = col;
        int idx = y * _img.width + x;
        for (int yy = 0; yy < h; yy++) {
            System.arraycopy(row, 0, _img.pixels, idx, w);
            idx += _img.width;
        }
        row = null;
    }
}



 
/*
 *    This work is licensed under a 
 *    Creative Commons Attribution-Noncommercial-Share Alike 3.0 License.
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/
 */
