
/*    Anymails
 *
 *        http://carohorn.de/anymails/
 *
 *    Concept, design, anything but code:
 *
 *        Carolin Horn, http://carohorn.de/
 *
 *    --------------------------------------------------------------------
 *        Code: Florian Jenett, http://florianjenett.de/
 *    --------------------------------------------------------------------
 *        Advisor: Brian Lucid
 *        Institute: Digital Media Institute Boston http://dmiboston.org/
 *    --------------------------------------------------------------------
 */
 
/*
 *    This work is licensed under a 
 *    Creative Commons Attribution-Noncommercial-Share Alike 3.0 License.
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/
 */
 
/*
 *    HOW TO RUN IT: please read the readme.txt
 */


import java.lang.reflect.*;
import processing.opengl.*;
// see: http://bezier.de/maus/
import de.bezier.maus.*;
import de.bezier.applemail.*;


int width2, height2; // width, height devided by 2

String version = "Anymails - Prototype 011 - fjenett 20070808";

final static boolean FULLSCREEN = false;

int initState = 0; // holdes the state of the initialization

AppleMailReader applemailreader; // loads and holds the whole apple mail database

Emlx[] totalMails;    // received and not deleted emails
Emlx[] mailsMatched;  // sorted and grouped subset, depends on settings in the main menu

Anymail[] anymails;        // all anymails
Anymail selectedAnymail;   // currently has mouse-focus
int totalGroups = 0;       // a counter to keep track of the number of groups currently displayed

AnymailComperator comperator;  // used for grouping and sorting the Emlx[] array
Mappable anymailToEmailId;     // used for reverse lookups: EmailID -> Anymail and opposite

// lot's a switches set by the menu
boolean[] speciesFilter;  // show / hide species
boolean allowUnread = true, allowRead = true, allowAnswered = true, allowJunk = true;
boolean refreshAge = true, stateChanged = true, visibilityChanged = true;

Area area51;  // the area the anymails can move in, stupid name
LoadThread loader;
GridLookUp gridLookUp;
String loadMessage;

long last_date_selection_start = -1;  // to see if it changed since last frame ..
long last_date_selection_end = -1;

boolean saveFrame = false;
boolean DEBUG = false;

void setup ()
{
    System.out.println( version );
    
    if ( FULLSCREEN )
        size( 1280, 854, OPENGL );
    else
        size( 1024, 768, OPENGL );
    
    if ( this.frame != null ) this.frame.setTitle( version );
    
    width2  = width/2;
    height2 = height/2;
    
    loadFonts();
    textFont( REGULAR_9 );
    
    loader = new LoadThread();
        
    applemailreader = AppleMailReader.getInstance();
    applemailreader.addListener( this ); // callback -> mailboxesLoaded(), see far below
    applemailreader.loadMailboxes( );

    background( Colors.BEIGE_MEDIUM );
    
    frameRate( 20 );
    
    comperator = new AnymailComperator();
}

boolean refreshMenu = true;

void draw ()
{
    long ts = System.currentTimeMillis();
    if ( initState == 0 )
    {
        // while applemailreader starts up ...
        loading();   
    }
    if ( initState > 0 && initState < 4 )
    {
        // while the loading thread is running ...
        if ( !loader.running ) loader.start();
        loading();
    }
    if ( initState == 4 )
    {
        // finally Anymails!
        anymail();
        drawMenu();
    }
    /*if ( saveFrame )
    {
        saveFrame("anymails_"+System.currentTimeMillis()+".tif"); 
        saveFrame = false;
    }*/
}

int loading_counter = 0;
PImage loadingimg;
void loading ()
{
    String lfile = "_data/loading/loading"+nf(loading_counter+1,4)+".png";
    loadingimg = loadImage(lfile);
    if ( loadingimg != null ) image(loadingimg, width2-(loadingimg.width/2), height2-(loadingimg.height/2));
    loading_counter++;
    loading_counter%=100;
}

void keyReleased ()
{
    /*if ( key == 's' )
        saveFrame = true;
    else if ( key == 'd' )
        DEBUG = !DEBUG;*/
}

void drawMenu ()
{
    if ( refreshMenu ) // just run once at startup since we can't draw in the loader thread
    {
        int mmX = (width  - mainmenuWidth) / 2;
        image( mainmenubackL, mmX, mainMenuY );
        image( mainmenubackC, mmX+mainmenubackL.width, mainMenuY );
        image( mainmenubackR, mmX+mainmenubackL.width+mainmenubackC.width, mainMenuY );
        refreshMenu = false;
        
        anymailsMenuContainer.disable();
        groupingMenuContainer.disable();
        timeMenuContainer.enable();
    }
    
    // if you are scrolling down and this is the first time you see
    // this fast looping monster, be prepared, i used it quite often ...
    // the idea is to let an exception happen and ignore it instead of
    // checking if "i >= 0 && i < elements.length".
    // another optimization is the "do { } while (..)" and stepping towards 0 (i--)
    // instead of upwards (i++)
    if ( iElements != null ) { int i=iElements.length-1; try { do {
        iElements[i].draw();    
    i--; } while(true); } catch (ArrayIndexOutOfBoundsException aioobe) { } }   
}

void anymail ()
{    
    // i used to have code here to clear each element seperatly (to gain speed),
    // but it turns out just drawing the backcolor over them is way faster!
    if ( area51 != null )
        area51.draw( Colors.BACKGROUND );
    
    long date_start = calendarslider.date_selection_start, date_end = calendarslider.date_selection_end;
    
    if (    last_date_selection_start != date_start
         || last_date_selection_end   != date_end
       )
    {
        mailsMatched =
            applemailreader.getFilteredMessages( 
                new EmlxFilter( date_start/(long)1000, // date from
                                date_end  /(long)1000, // date to
                                EmlxFilter.SKIP, //(allowRead ? EmlxFilter.READ : EmlxFilter.SKIP ), // read
                                EmlxFilter.SKIP, //(allowAnswered ? EmlxFilter.ANSWERED : EmlxFilter.SKIP ), // answered
                                EmlxFilter.SKIP, // sender id
                                EmlxFilter.SKIP ), // mailbox id
                                EmlxFilter.RECEIVED_ONLY,
                                EmlxFilter.HIDE_DELETED );
        
        refreshAge = true;
        visibilityChanged = true;
        stateChanged = true;
        selectedAnymail = null;
    }
    
    if ( visibilityChanged && mailsMatched != null && mailsMatched.length > 0 )
    {
        mm.removeMausManager( area51 );
        
        area51 = new Area( 0, 0, width, mainMenuY-5 );
        area51.singleMode = true;
        
        mm.registerMausManager( area51 );
                                
            int i=mailsMatched.length-1; try { do {
                if ( checkMailVisible(mailsMatched[i]) ) {
                    anymails[anymailToEmailId.indexForValue(mailsMatched[i].id)].setArea(area51);
                }
            i--; } while(true); } catch (ArrayIndexOutOfBoundsException aioobe) { }
        
        visibilityChanged = false;
    }
    
    /*if ( DEBUG )
    {
        stroke(Colors.BEIGE_DARK);
        strokeWeight( 1 );
        for ( int x = 0; x < gridLookUp.wGrid; x++ ) line(x*gridLookUp.gridSizeP,0, x*gridLookUp.gridSizeP,mainMenuY);
        for ( int y = 0; y < gridLookUp.hGrid; y++ ) line(0,y*gridLookUp.gridSizeP, width,y*gridLookUp.gridSizeP);
        noStroke();
    }*/
    
    if ( stateChanged && mailsMatched != null && mailsMatched.length > 0 )
    {
        Arrays.sort(mailsMatched, comperator);
        Anymail lastTemp = null;
        totalGroups = 0;
        
        int i=0; try { do {
            Anymail tmp = anymails[anymailToEmailId.indexForValue(mailsMatched[i].id)];
            tmp.unChain();
            if ( checkMailVisible(mailsMatched[i]) )
            {
                if ( refreshAge )
                    tmp.setAge(date_start,date_end);
                    
                tmp.setHighlight( false, Anymail.THIS_ONLY );
                        
                if ( lastTemp != null && tmp.compareGrouping(lastTemp) == 0 )
                {
                    tmp.chain( lastTemp );
                }
                else
                    totalGroups++;
                
                tmp.groupID = totalGroups;
                        
                lastTemp = tmp;
            }
        i++; } while( true ); } catch (ArrayIndexOutOfBoundsException aioobe) { }
        
        stateChanged = false;
    }
    
    last_date_selection_start = date_start;
    last_date_selection_end = date_end;
    
    /*if ( DEBUG )
        debugFeld(date_start,date_end);*/
    
    gridLookUp.clear();
      
    if ( mailsMatched != null && mailsMatched.length > 0 )
    {            
        int i;
        
        i=mailsMatched.length-1; try { do {
            if ( checkMailVisible(mailsMatched[i]) ) {
                Anymail tmp = anymails[anymailToEmailId.indexForValue(mailsMatched[i].id)];
                if ( refreshAge )
                    tmp.setAge(date_start,date_end);
                tmp.move();
                gridLookUp.add(tmp);
                
                if ( Grouping.WHEN_STATE == Grouping.WHEN_CONTINOUSLY )
                {
                    if (!tmp.highLight)
                        tmp.setHighlight( true, Anymail.THIS_ONLY );
                    tmp.drawHighlight(Anymail.THIS_ONLY, Anymail.DRAW);
                }
            }
        i--; } while(true); } catch (ArrayIndexOutOfBoundsException aioobe) { }
        
        if ( Grouping.WHEN_STATE == Grouping.WHEN_ROLLOVER 
                  && selectedAnymail != null )
        {
            selectedAnymail.drawHighlight();
        }
        
        i=mailsMatched.length-1; try { do {
            if ( checkMailVisible(mailsMatched[i]) ) {
                anymails[anymailToEmailId.indexForValue(mailsMatched[i].id)].draw();
           }
        i--; } while(true); } catch (ArrayIndexOutOfBoundsException aioobe) { }
        
        if ( selectedAnymail != null )
            selectedAnymail.drawInfo();
    }
    
    refreshAge = false;
}

boolean checkMailVisible ( Emlx amail )
{
    return (  (   (allowUnread   && !amail.read)
               || (allowRead     && amail.read    && !amail.answered)
               || (allowAnswered && amail.answered)
               )
             && (allowJunk     || !amail.is_junk)
             && speciesFilter[anymails[anymailToEmailId.indexForValue(amail.id)].type] );
}

void debugFeld (long date_end, long date_start) {
        if ( true ) return;
        noStroke();
        //fill( Colors.BEIGE_MEDIUM );
        //rect( 0,0, 300, 200 );
        fill( Colors.GREY_DARK );
        textFont( REGULAR_10 );
        textAlign( LEFT );
        text( (new Date(date_start)).toString(), 20,20 );
        text( (new Date(date_end)).toString(), 20,30 );
        text( "Found " + (mailsMatched != null ? mailsMatched.length : 0) + " emails of " + totalMails.length, 20,40 );
        if ( mailsMatched != null && mailsMatched.length > 0 )
        {
            int ur=0, re=0, ra=0, ju=0, de=0;
            for ( int ii = 0; ii < mailsMatched.length; ii++ )
            {
                if ( mailsMatched[ii].deleted
                     || applemailreader.mailboxes[ applemailreader.mailboxMap.indexForValue( mailsMatched[ii].mailbox_id ) ].isTrash) { 
                     //println( applemailreader.mailboxes[ Mailbox.key( mailsMatched[ii].mailbox_id ) ].folderpath );
                     de++; continue; }
                     
                if ( allowUnread && !mailsMatched[ii].read ) { ur++; }
                if ( allowRead && mailsMatched[ii].read && !mailsMatched[ii].answered ) { re++; }
                if ( allowAnswered && mailsMatched[ii].answered ) { ra++; }
                if ( allowJunk && mailsMatched[ii].is_junk ) { ju++; }
                //println( applemailreader.mailboxes[ Mailbox.key( mailsMatched[ii].mailbox_id ) ].folderpath );
            }
            text( "\t" + ur + " new, " + re + " read, " + ra + " answered, " + ju + " junk." , 20,50 ); //  " + de + " deleted."
        }
        text( "Current framerate: " + frameRate , 20, 60 ); //  " + de + " deleted."
}
